from dateutil import parser


def parse_isoformat(date_string):
    """ Parses the current date format version """
    if date_string:
        return parser.isoparse(date_string)
    else:
        return None



# This is duplicated from AI. Sharing these somewhere would be helpful...

simple_query_string_schema = {
    'properties': {
        'query': {
            'type': 'object',
            'properties': {
                'simple_query_string': {
                    'type': 'object',
                    'properties': {
                        'query': {
                            'type': 'string'
                        },
                        'fields': {
                            'type': 'array',
                            'items': {
                                'type': 'string'
                            }
                        },
                        'default_operator': {
                            'type': 'string'
                        }
                    },
                    'additionalProperties': False,
                    'required': ['query', 'fields'],
                }
            },
            'additionalProperties': False,
            'required': ['simple_query_string'],
        }
    },
    'additionalProperties': False,
    'required': ['query']
}

query_string_schema = {
    'properties': {
        'query': {
            'type': 'object',
            'properties': {
                'query_string': {
                    'type': 'object',
                    'properties': {
                        'default_field': {
                            'type': 'string'
                        },
                        'query': {
                            'type': 'string'
                        }
                    },
                    'additionalProperties': False,
                    'required': ['query']
                }
            },
            'additionalProperties': False,
            'required': ['query_string']
        }
    },
    'additionalProperties': False,
    'required': ['query']
}

exists_schema = {
    'properties': {
        'query': {
            'type': 'object',
            'properties': {
                'exists': {
                    'type': 'object',
                    'properties': {
                        'field': {
                            'type': 'string'
                        }
                    },
                    'additionalProperties': False,
                    'required': ['field']
                },
            },
            'additionalProperties': False,
            'required': ['exists']
        }
    },
    'additionalProperties': False,
    'required': ['query']
}

range_schema = {
    'properties': {
        'query': {
            'type': 'object',
            'properties': {
                'range': {
                    'type': 'object',
                    'patternProperties': {
                        '^.*$': {
                            'type': 'object',
                            'patternProperties': {
                                'gte?': {
                                    "anyOf": [
                                        {"type": "string"},
                                        {"type": "number"}
                                    ]
                                },
                                'lte?': {
                                    'anyOf': [
                                        {"type": "string"},
                                        {"type": "number"}
                                    ]
                                },
                                'boost': {
                                    'type': 'number'
                                }

                            },
                            'additionalProperties': False
                        }
                    },
                    'additionalProperties': False
                },
            },
            'additionalProperties': False,
            'required': ['range']
        }
    },
    'additionalProperties': False,
    'required': ['query']
}

terms_query = {
    'properties': {
        'query': {
            'type': 'object',
            'properties': {
                'terms': {
                    'type': 'object',
                    'patternProperties': {
                        '^.*$': {
                            'type': 'array',
                            'items': {
                                'anyOf': [
                                    {"type": "string"},
                                    {"type": "number"},
                                    {"type": "integer"}
                                ]
                            }
                        }
                    },
                    'additionalProperties': False
                }
            },
            'additionalProperties': False,
            'required': ['terms']

        }
    },
    'additionalProperties': False,
    'required': ['query']

}

search_schema = {
    'type': 'object',
    'oneOf': [
        simple_query_string_schema,
        query_string_schema,
        exists_schema,
        range_schema,
        terms_query
    ]
}



import logging
import requests
from urllib3.util.retry import Retry
from requests.adapters import HTTPAdapter

from . import const

logger = logging.getLogger(__name__)


class PolyswarmHTTP(requests.Session):
    def __init__(self, key, retries, user_agent=const.DEFAULT_USER_AGENT):
        super(PolyswarmHTTP, self).__init__()
        logger.debug('Creating PolyswarmHTTP instance')
        self.requests_retry_session(retries=retries)

        if key:
            self.set_auth(key)

        if user_agent:
            self.set_user_agent(user_agent)

    def requests_retry_session(self, retries=const.DEFAULT_RETRIES, backoff_factor=const.DEFAULT_BACKOFF,
                               status_forcelist=const.DEFAULT_RETRY_CODES):
        retry = Retry(
            total=retries,
            read=retries,
            connect=retries,
            backoff_factor=backoff_factor,
            status_forcelist=status_forcelist,
        )
        adapter = HTTPAdapter(max_retries=retry)
        self.mount('http://', adapter)
        self.mount('https://', adapter)

    def set_auth(self, key):
        if key:
            self.headers.update({'Authorization': key})
        else:
            self.headers.pop('Authorization', None)

    def set_user_agent(self, ua):
        if ua:
            self.headers.update({'User-Agent': ua})
        else:
            self.headers.pop('User-Agent', None)



import logging

from jsonschema import validate, ValidationError

from .. import exceptions

logger = logging.getLogger(__name__)


class BasePSType(object):
    def __init__(self, polyswarm=None):
        self.polyswarm = polyswarm


class BasePSResourceType(BasePSType):
    @classmethod
    def parse_result(cls, api_instance, result, **kwargs):
        logger.debug('Parsing resource %s', cls.__name__)
        return cls(result, polyswarm=api_instance, **kwargs)

    @classmethod
    def parse_result_list(cls, api_instance, json_data, **kwargs):
        return [cls.parse_result(api_instance, entry, **kwargs) for entry in json_data]


class BasePSJSONType(BasePSResourceType):
    SCHEMA = {
        'type': ['object', 'array']
    }

    def __init__(self, json=None, polyswarm=None):
        super(BasePSJSONType, self).__init__(polyswarm=polyswarm)
        self._json = None
        if json is not None:
            self.json = json

    @property
    def json(self):
        return self._json

    @json.setter
    def json(self, value):
        # this is expensive on thousands of objects
        # avoid if disabled
        if self.polyswarm and self.polyswarm.validate:
            self._validate(value)
        self._json = value

    def _validate(self, json, schema=None):
        if not schema:
            schema = self.SCHEMA

        try:
            validate(json, schema)
        except ValidationError:
            raise exceptions.InvalidJSONResponseException("Failed to validate json against schema", json, self.SCHEMA)


# TODO better way to do this with ABC?
class Hashable:
    @property
    def hash(self):
        return self.sha256

    @property
    def hash_type(self):
        return 'sha256'

    def __eq__(self, other):
        return self.hash == other


class AsInteger:
    def __int__(self):
        return int(self.id)



import logging
import os
import io
import functools
from binascii import unhexlify
from enum import Enum
from hashlib import sha256 as _sha256, sha1 as _sha1, md5 as _md5

from future.utils import raise_from, string_types

from polyswarm_api.const import FILE_CHUNK_SIZE

try:
    import yara
except ImportError:
    yara = None

from polyswarm_api import exceptions
from polyswarm_api import const
from . import base
from . import schemas
from . import date

logger = logging.getLogger(__name__)


#####################################################################
# Resources returned by the API
#####################################################################

class Engine(base.BasePSJSONType):
    def __init__(self, json, polyswarm=None):
        super(Engine, self).__init__(json=json, polyswarm=polyswarm)
        self.address = json['address'].lower()
        self.name = json.get('name')


class Metadata(base.BasePSJSONType, base.AsInteger):
    KNOWN_KEYS = {'artifact', 'exiftool', 'hash', 'lief', 'pefile', 'scan', 'strings'}

    def __init__(self, json, polyswarm=None):
        super(Metadata, self).__init__(json=json, polyswarm=polyswarm)
        self.created = date.parse_isoformat(self.artifact.get('created'))

        self.id = self.artifact.get('id')

        self.sha1 = self.hash.get('sha1')
        self.sha256 = self.hash.get('sha256')
        self.md5 = self.hash.get('md5')
        self.ssdeep = self.hash.get('ssdeep')
        self.tlsh = self.hash.get('tlsh')

        self.first_seen = date.parse_isoformat(self.scan.get('first_seen'))
        self.last_seen = date.parse_isoformat(self.scan.get('last_seen'))
        self.mimetype = self.scan.get('mimetype', {}).get('mime')
        self.extended_mimetype = self.scan.get('mimetype', {}).get('extended')
        self.detections = self.scan.get('detections', {}).get('malicious')
        self.total_detections = self.scan.get('detections', {}).get('total')

        self.domains = self.strings.get('domains')
        self.ipv4 = self.strings.get('ipv4')
        self.ipv6 = self.strings.get('ipv6')
        self.urls = self.strings.get('urls')

    def __contains__(self, item):
        return item in self.json

    def __getattr__(self, name):
        try:
            return self.json[name]
        except KeyError:
            if name in Metadata.KNOWN_KEYS:
                return {}
            raise AttributeError()


class ArtifactInstance(base.BasePSJSONType, base.Hashable, base.AsInteger):
    SCHEMA = schemas.artifact_instance_schema

    def __init__(self, json, polyswarm=None):
        super(ArtifactInstance, self).__init__(json=json, polyswarm=polyswarm)
        self.id = json['id']
        self.artifact_id = json['artifact_id']
        self.assertions = [Assertion(self, a, polyswarm) for a in json['assertions']]
        self.country = json['country']
        self.community = json['community']
        self.created = date.parse_isoformat(json['created'])
        self.extended_type = json['extended_type']
        self.failed = json['failed']
        self.filename = json['filename']
        self.last_seen = date.parse_isoformat(json['last_seen'])
        self.first_seen = date.parse_isoformat(json['first_seen'])
        self.md5 = json['md5']
        self.mimetype = json['mimetype']
        self.result = json['result']
        self.sha1 = json['sha1']
        self.sha256 = json['sha256']
        self.size = json['size']
        self.type = json['type']
        self.votes = [Vote(self, v, polyswarm) for v in json['votes']]
        self.window_closed = json['window_closed']
        self.polyscore = float(json['polyscore']) if json.get('polyscore') is not None else None
        self.permalink = const.DEFAULT_PERMALINK_BASE + '/' + str(self.hash)

        metadata_json = json.get('metadata') or []
        metadata = {metadata['tool']: metadata['tool_metadata'] for metadata in metadata_json}
        self.metadata = Metadata(metadata, polyswarm)

        self._detections = None
        self._valid_assertions = None
        self._filenames = None

    def __str__(self):
        return "ArtifactInstance-<%s>" % self.hash

    @property
    def detections(self):
        if not self._detections:
            self._detections = [a for a in self.assertions if a.mask and a.verdict]
        return self._detections

    @property
    def valid_assertions(self):
        if not self._valid_assertions:
            self._valid_assertions = [a for a in self.assertions if a.mask]
        return self._valid_assertions

    @property
    def filenames(self):
        if self._filenames is None:
            for metadata in self.json.get('metadata', []):
                if metadata.get('tool') == 'scan':
                    self._filenames = metadata.get('tool_metadata', {}).get('filename', [])
                    break
            else:
                self._filenames = []
        return self._filenames


class ArtifactArchive(base.BasePSJSONType, base.AsInteger):
    SCHEMA = schemas.artifact_archive_schema

    def __init__(self, json, polyswarm=None):
        super(ArtifactArchive, self).__init__(json=json, polyswarm=polyswarm)
        self.id = json['id']
        self.community = json['community']
        self.created = date.parse_isoformat(json['created'])
        self.uri = json['uri']


class Hunt(base.BasePSJSONType, base.AsInteger):
    SCHEMA = schemas.hunt_status

    def __init__(self, json, polyswarm=None):
        super(Hunt, self).__init__(json=json, polyswarm=polyswarm)
        # active only present for live hunts
        self.id = json['id']
        self.created = date.parse_isoformat(json['created'])
        self.status = json['status']
        self.active = json.get('active')
        self.ruleset_name = json.get('ruleset_name')


class HuntResult(base.BasePSJSONType, base.AsInteger):
    SCHEMA = schemas.hunt_result

    def __init__(self, json, polyswarm=None):
        super(HuntResult, self).__init__(json=json, polyswarm=polyswarm)
        self.id = json['id']
        self.rule_name = json['rule_name']
        self.tags = json['tags']
        self.created = date.parse_isoformat(json['created'])
        self.sha256 = json['sha256']
        self.historicalscan_id = json['historicalscan_id']
        self.livescan_id = json['livescan_id']
        self.artifact = ArtifactInstance(json['artifact'], polyswarm)


def _read_chunks(file_handle):
    while True:
        data = file_handle.read(FILE_CHUNK_SIZE)
        if not data:
            break
        yield data


def all_hashes(file_handle, algorithms=(_sha256, _sha1, _md5)):
    hashers = [alg() for alg in algorithms]
    for data in _read_chunks(file_handle):
        [h.update(data) for h in hashers]
    return [Hash(h.hexdigest()) for h in hashers]


class LocalHandle(base.BasePSResourceType):
    def __init__(self, contents, polyswarm=None, handle=None):
        super(LocalHandle, self).__init__(polyswarm=polyswarm)
        self.handle = handle or io.BytesIO()
        for chunk in contents:
            self.handle.write(chunk)

    # Inspired by
    # https://github.com/python/cpython/blob/29500737d45cbca9604d9ce845fb2acc3f531401/Lib/tempfile.py#L461
    def __getattr__(self, name):
        # Attribute lookups are delegated to the underlying file
        # and cached for non-numeric results
        # (i.e. methods are cached, closed and friends are not)
        a = getattr(self.handle, name)
        if hasattr(a, '__call__'):
            func = a

            @functools.wraps(func)
            def func_wrapper(*args, **kwargs):
                return func(*args, **kwargs)

            a = func_wrapper
        if not isinstance(a, int):
            setattr(self, name, a)
        return a


class LocalArtifact(LocalHandle, base.Hashable):
    """ Artifact for which we have local content """
    def __init__(self, handle, artifact_name=None, artifact_type=None, polyswarm=None, analyze=True):
        """
        A representation of an artifact we have locally

        :param artifact_name: Name of the artifact
        :param artifact_type: Type of artifact
        :param polyswarm: PolyswarmAPI instance
        :param analyze: Boolean, if True will run analyses on artifact on startup (Note: this may still run later if False)
        """
        # create the LocalHandle with the given handle and don't write anything to it
        super(LocalArtifact, self).__init__(b'', polyswarm=polyswarm, handle=handle)
        self.artifact_type = artifact_type or ArtifactType.FILE
        self.artifact_name = artifact_name or self.hash

        self.sha256 = None
        self.sha1 = None
        self.md5 = None
        self.analyzed = False
        if analyze:
            self.analyze_artifact()

    @classmethod
    def from_path(cls, api, path, artifact_type=None, analyze=False, create=False, **kwargs):
        if not isinstance(path, string_types):
            raise exceptions.InvalidValueException('Path should be a string')
        folder, file_name = os.path.split(path)
        if create:
            # TODO: this should be replaced with os.makedirs(path, exist_ok=True)
            #  once we drop support to python 2.7
            if not os.path.exists(folder):
                os.makedirs(folder)
        elif not os.path.isfile(path):
            raise exceptions.ArtifactDeletedException("The file does not exist")

        mode = kwargs.pop('mode', 'wb+' if create else 'rb')
        handler = open(path, mode=mode, **kwargs)
        return cls(handler, artifact_name=file_name, artifact_type=artifact_type, analyze=analyze, polyswarm=api)

    @classmethod
    def from_content(cls, api, content, artifact_name=None, artifact_type=None, analyze=False):
        if isinstance(content, string_types):
            content = content.encode("utf8")
        handler = io.BytesIO(content)
        return cls(handler, artifact_name=artifact_name, artifact_type=artifact_type, analyze=analyze, polyswarm=api)

    @property
    def hash(self):
        self.analyze_artifact()
        return super(LocalArtifact, self).hash

    def analyze_artifact(self, force=False):
        if not self.analyzed or force:
            self.handle.seek(0)
            self._calc_hashes(self.handle)
            self.handle.seek(0)
            self._run_analyzers(self.handle)
            self.analyzed = True

    def _calc_hashes(self, fh):
        self.sha256, self.sha1, self.md5 = all_hashes(fh)

    def _run_analyzers(self, fh):
        # TODO implement custom analyzer support, so users can implement plugins here.
        return {}

    def __str__(self):
        return "Artifact <%s>" % self.hash


class YaraRuleset(base.BasePSJSONType, base.AsInteger):
    def __init__(self, json, polyswarm=None):
        super(YaraRuleset, self).__init__(json, polyswarm)
        self.yara = json['yara']
        self.name = json.get('name')
        self.id = json.get('id')
        self.description = json.get('description')
        self.created = date.parse_isoformat(json.get('created'))
        self.modified = date.parse_isoformat(json.get('modified'))
        self.deleted = json.get('deleted')

        if not self.yara:
            raise exceptions.InvalidValueException("Must provide yara ruleset content")

    def validate(self):
        if not yara:
            raise exceptions.NotImportedException("Cannot validate rules locally without yara-python")

        try:
            yara.compile(source=self.yara)
        except yara.SyntaxError as e:
            raise exceptions.InvalidYaraRulesException('Malformed yara file: {}'.format(e.args[0]) + '\n')

        return True


class TagLink(base.BasePSJSONType, base.AsInteger):
    def __init__(self, json, polyswarm=None):
        super(TagLink, self).__init__(json, polyswarm)
        self.id = json.get('id')
        self.sha256 = json.get('sha256')
        self.created = date.parse_isoformat(json.get('created'))
        self.updated = date.parse_isoformat(json.get('updated'))
        self.first_seen = date.parse_isoformat(json.get('first_seen'))
        self.tags = json.get('tags')
        self.families = json.get('families')


class MalwareFamily(base.BasePSJSONType, base.AsInteger):
    def __init__(self, json, polyswarm=None):
        super(MalwareFamily, self).__init__(json, polyswarm)
        self.id = json.get('id')
        self.created = date.parse_isoformat(json.get('created'))
        self.updated = date.parse_isoformat(json.get('updated'))
        self.name = json.get('name')
        self.emerging = date.parse_isoformat(json.get('emerging'))


class Tag(base.BasePSJSONType, base.AsInteger):
    def __init__(self, json, polyswarm=None):
        super(Tag, self).__init__(json, polyswarm)
        self.id = json.get('id')
        self.created = date.parse_isoformat(json.get('created'))
        self.updated = date.parse_isoformat(json.get('updated'))
        self.name = json.get('name')


#####################################################################
# Nested Resources
#####################################################################


class Assertion(base.BasePSJSONType):
    SCHEMA = schemas.assertion_schema

    def __init__(self, scanfile, json, polyswarm=None):
        super(Assertion, self).__init__(json=json, polyswarm=polyswarm)
        self.scanfile = scanfile
        self.author = json['author']
        self.author_name = json['author_name']
        self.engine_name = json['engine'].get('name')
        self.bid = int(json['bid'])
        self.mask = json['mask']
        # deal with metadata being a string instead of null
        self.metadata = json['metadata'] if json['metadata'] else {}
        self.verdict = json['verdict']

    def __str__(self):
        return "Assertion-%s: %s" % (self.engine_name, self.verdict)


class Vote(base.BasePSJSONType):
    SCHEMA = schemas.vote_schema

    def __init__(self, scanfile, json, polyswarm=None):
        super(Vote, self).__init__(json=json, polyswarm=polyswarm)
        self.scanfile = scanfile
        self.arbiter = json['arbiter']
        self.vote = json['vote']

    def __str__(self):
        return "Vote-%s: %s" % (self.arbiter, self.vote)


#####################################################################
# Resources Used as input parameters in PolyswarmAPI
#####################################################################


class ArtifactType(Enum):
    FILE = 0
    URL = 1

    @staticmethod
    def parse(value):
        if isinstance(value, ArtifactType):
            return value
        try:
            return ArtifactType[value.upper()]
        except Exception as e:
            raise raise_from(
                exceptions.InvalidValueException(
                    'Unable to get the artifact type from the provided value {}'.format(value)), e)

    @staticmethod
    def to_string(artifact_type):
        return artifact_type.name.lower()

    def decode_content(self, content):
        if content is None:
            return None

        if self == ArtifactType.URL:
            try:
                return content.decode('utf-8')
            except UnicodeDecodeError:
                raise exceptions.DecodeErrorException('Error decoding URL')
        else:
            return content


def is_hex(value):
    try:
        _ = int(value, 16)
        return True
    except ValueError:
        return False


def is_valid_sha1(value):
    if len(value) != 40:
        return False
    return is_hex(value)


def is_valid_md5(value):
    if len(value) != 32:
        return False
    return is_hex(value)


def is_valid_sha256(value):
    if len(value) != 64:
        return False
    return is_hex(value)


class Hash(base.Hashable, base.BasePSType):
    SCHEMA = {'type': ['string', 'null']}

    SUPPORTED_HASH_TYPES = {
        'sha1': is_valid_sha1,
        'sha256': is_valid_sha256,
        'md5': is_valid_md5,
    }

    def __init__(self, hash_, hash_type=None, polyswarm=None):
        super(Hash, self).__init__(polyswarm=polyswarm)
        hash_ = hash_.strip()

        if hash_type and hash_type not in Hash.SUPPORTED_HASH_TYPES:
            raise exceptions.InvalidValueException('Hash type provided is not supported.')

        self._hash_type = Hash.get_hash_type(hash_)

        if self._hash_type is None:
            raise exceptions.InvalidValueException('Invalid hash provided: {}'.format(hash_))

        if hash_type and self.hash_type != hash_type:
            raise exceptions.InvalidValueException('Detected hash type {}, got {} for hash {}'
                                                   .format(hash_type, self.hash_type, hash_))

        self._hash = hash_

    @classmethod
    def from_hashable(cls, hash_, polyswarm=None, hash_type=None):
        """
        Coerce to Hashable object

        :param hash_: Hashable object
        :param hash_type: Hash type
        :param polyswarm: PolyswarmAPI instance
        :return: Hash
        """
        if issubclass(type(hash_), base.Hashable):
            if hash_type and hash_.hash_type != hash_type:
                raise exceptions.InvalidValueException('Detected hash type {}, got {} for hashable {}'
                                                       .format(hash_.hash_type, hash_type, hash_.hash))
            return hash_
        return Hash(hash_, hash_type=hash_type, polyswarm=polyswarm)

    @classmethod
    def get_hash_type(cls, value):
        for hash_type, check in cls.SUPPORTED_HASH_TYPES.items():
            if check(value):
                return hash_type
        return None

    @property
    def raw(self):
        return unhexlify(self.hash)

    def search(self):
        if not self.polyswarm:
            raise exceptions.MissingAPIInstanceException("Missing API instance")
        return self.polyswarm.search_hashes([self])

    @property
    def hash(self):
        return self._hash

    @property
    def hash_type(self):
        return self._hash_type

    def __hash__(self):
        return hash(self.hash)

    def __str__(self):
        return self.hash

    def __repr__(self):
        return "{}={}".format(self.hash_type, self.hash)



hunt_submission = {
    'type': 'object',
    'properties': {
        'hunt_id': {'type': 'string'},
    },
    'required': ['hunt_id'],
}

hunt_status = {
    'type': 'object',
    'properties': {
        'active': {'type': 'boolean'},
        'created': {'type': 'string'},
        'id': {'type': 'string'},
        'results': {'type': 'array'},
        'total': {'type': 'integer'},
        'status': {'type': 'string'},
    },
    'required': ['created', 'id', 'results', 'status']
}

hunt_result = {
    'type': 'object',
    'properties': {
        'artifact': {'type': 'object'},
        'rule_name': {'type': 'string'},
        'tags': {'type': 'string'},
    }
}




import os
import platform
from . import _version

# API constants
DEFAULT_GLOBAL_API = 'https://api.polyswarm.network/v2'
DEFAULT_PERMALINK_BASE = os.getenv('POLYSWARM_PORTAL_URI', 'https://polyswarm.network/scan/results/file')
DEFAULT_COMMUNITY = 'default'
DEFAULT_SCAN_TIMEOUT = 60*15
RESULT_CHUNK_SIZE = 100
POLL_FREQUENCY = 1

# HTTP settings
DEFAULT_HTTP_TIMEOUT = 30
DEFAULT_RETRIES = 0
DEFAULT_BACKOFF = 1
DEFAULT_RETRY_CODES = (502, 504)
DEFAULT_USER_AGENT = 'polyswarm-api/{} ({}-{}-{}-{})'.format(
    _version.__version__, platform.machine(), platform.system(),
    platform.python_implementation(), platform.python_version(),
)

# concurrent HTTP workers
DEFAULT_WORKER_COUNT = 8

# API maximums
MAX_HUNT_RESULTS = 20000
MAX_ARTIFACT_BATCH_SIZE = 256

# Filesystem constants
FILE_CHUNK_SIZE = 8192
MAX_OPEN_FDS = 256
# this results in worst case 32MB memory usage during downloads
DOWNLOAD_CHUNK_SIZE = 1024*1024*4

MAX_SINCE_TIME_STREAM = 2 * 60 * 24

USAGE_EXCEEDED_MESSAGE = 'Usage limits were exceeded. This may mean you need to purchase a ' \
                         'larger package, or that you have exceeded rate limits.\n' \
                         'If you continue to have issues, please contact us at info@polyswarm.io.'




import logging
import json
from future.utils import raise_from
from copy import deepcopy

from . import const
from . import http
from . import exceptions
from .types import resources

try:
    from json.decoder import JSONDecodeError
except ImportError:
    JSONDecodeError = ValueError


logger = logging.getLogger(__name__)


class RequestParamsEncoder(json.JSONEncoder):
    def default(self, obj):
        try:
            return json.JSONEncoder.default(self, obj)
        except Exception:
            return str(obj)


class PolyswarmRequest(object):
    """This class holds a requests-compatible dictionary and extra information we need to parse the response."""
    def __init__(self, api_instance, request_parameters, key=None, result_parser=None, json_response=True, **kwargs):
        logger.debug('Creating PolyswarmRequest instance.\nRequest parameters: %s\nResult parser: %s',
                     request_parameters, result_parser.__name__)
        self.api_instance = api_instance
        # we should not access the api_instance session directly, but provide as a
        # parameter in the constructor, but this will do for the moment
        self.session = self.api_instance.session or http.PolyswarmHTTP(key, retries=const.DEFAULT_RETRIES)
        self.timeout = self.api_instance.timeout or const.DEFAULT_HTTP_TIMEOUT
        self.request_parameters = request_parameters
        self.result_parser = result_parser
        self.json_response = json_response
        self.raw_result = None
        self.status_code = None
        self.status = None
        self.result = None
        self.errors = None
        self.total = None
        self.limit = None
        self.offset = None
        self.order_by = None
        self.direction = None
        self.has_more = None
        self.parser_kwargs = kwargs

    def execute(self):
        logger.debug('Executing request.')
        self.request_parameters.setdefault('timeout', self.timeout)
        if not self.json_response:
            self.request_parameters.setdefault('stream', True)
        self.raw_result = self.session.request(**self.request_parameters)
        logger.debug('Request returned code %s with content:\n%s',
                     self.raw_result.status_code, self.raw_result.content)
        if self.result_parser is not None:
            self.parse_result(self.raw_result)
        return self

    def _bad_status_message(self):
        request_parameters = json.dumps(self.request_parameters, indent=4, sort_keys=True, cls=RequestParamsEncoder)
        message = "Error when running the request:\n{}\n" \
                  "Return code: {}\n" \
                  "Message: {}".format(request_parameters,
                                       self.status_code,
                                       self.result)
        if self.errors:
            message = '{}\nErrors:\n{}'.format(message, '\n'.join(str(error) for error in self.errors))
        return message

    def _extract_json_body(self, result):
        self.json = result.json()
        self.result = self.json.get('result')
        self.status = self.json.get('status')
        self.errors = self.json.get('errors')

    def parse_result(self, result):
        logger.debug('Parsing request results.')
        self.status_code = result.status_code
        try:
            if self.status_code // 100 != 2:
                self._extract_json_body(result)
                if self.status_code == 429:
                    raise exceptions.UsageLimitsExceededException(self, const.USAGE_EXCEEDED_MESSAGE)
                if self.status_code == 404:
                    raise exceptions.NotFoundException(self, self.result)
                raise exceptions.RequestException(self, self._bad_status_message())
            elif self.status_code == 204:
                raise exceptions.NoResultsException(self, 'The request returned no results.')
            elif self.json_response:
                self._extract_json_body(result)
                self.total = self.json.get('total')
                self.limit = self.json.get('limit')
                self.offset = self.json.get('offset')
                self.order_by = self.json.get('order_by')
                self.direction = self.json.get('direction')
                self.has_more = self.json.get('has_more')
                if 'result' in self.json:
                    result = self.json['result']
                elif 'results' in self.json:
                    result = self.json['results']
                else:
                    raise exceptions.RequestException(
                        self,
                        'The response standard must contain either the "result" or "results" key.'
                    )
                if isinstance(result, list):
                    self.result = self.result_parser.parse_result_list(self.api_instance, result, **self.parser_kwargs)
                else:
                    self.result = self.result_parser.parse_result(self.api_instance, result, **self.parser_kwargs)
            else:
                self.result = self.result_parser.parse_result(self.api_instance,
                                                              result.iter_content(const.DOWNLOAD_CHUNK_SIZE),
                                                              **self.parser_kwargs)
        except JSONDecodeError as e:
            if self.status_code == 404:
                raise raise_from(exceptions.NotFoundException(self, 'The requested endpoint does not exist.'), e)
            else:
                raise raise_from(exceptions.RequestException(self, 'Server returned non-JSON response.'), e)

    def __iter__(self):
        return self.consume_results()

    def consume_results(self):
        # StopIteration is deprecated
        # As per https://www.python.org/dev/peps/pep-0479/
        # We simply return upon termination condition
        request = self
        while True:
            # consume items items from list if iterable
            # of yield the single result if not
            try:
                for result in request.result:
                    yield result
            except TypeError:
                yield request.result
                # if the result is not a list, there is not next page
                return

            # if the server indicates that there are no more results, return
            if not request.has_more:
                return
            # try to get the next page and execute the request
            request = request.next_page().execute()

    def next_page(self):
        new_parameters = deepcopy(self.request_parameters)
        params = new_parameters.setdefault('params', {})
        if isinstance(params, dict):
            params['offset'] = self.offset
            params['limit'] = self.limit
        else:
            params = [p for p in params if p[0] != 'offset' and p[0] != 'limit']
            params.extend([('offset', self.offset), ('limit', self.limit)])
            new_parameters['params'] = params
        return PolyswarmRequest(
            self.api_instance,
            new_parameters,
            result_parser=self.result_parser,
        )


class PolyswarmRequestGenerator(object):
    """ This class will return PolyswarmRequests"""
    def __init__(self, api_instance):
        logger.debug('Creating PolyswarmRequestGenerator instance')
        self.api_instance = api_instance
        self.uri = api_instance.uri
        self.community = api_instance.community

    def download(self, hash_value, hash_type, handle=None):
        return PolyswarmRequest(
            self.api_instance,
            {
                'method': 'GET',
                'url': '{}/download/{}/{}'.format(self.uri, hash_type, hash_value),
                'stream': True,
            },
            json_response=False,
            result_parser=resources.LocalHandle,
            handle=handle,
        )

    def download_archive(self, u, handle=None):
        """ This method is special, in that it is simply for downloading from S3 """
        return PolyswarmRequest(
            self.api_instance,
            {
                'method': 'GET',
                'url': u,
                'stream': True,
                'headers': {'Authorization': None}
            },
            json_response=False,
            result_parser=resources.LocalHandle,
            handle=handle,
        )

    def stream(self, since=const.MAX_SINCE_TIME_STREAM):
        return PolyswarmRequest(
            self.api_instance,
            {
                'method': 'GET',
                'url': '{}/consumer/download/stream'.format(self.uri),
                'params': {'since': since},
            },
            result_parser=resources.ArtifactArchive,
        )

    def search_hash(self, hash_value, hash_type):
        return PolyswarmRequest(
            self.api_instance,
            {
                'method': 'GET',
                'url': '{}/search/hash/{}'.format(self.uri, hash_type),
                'params': {
                    'hash': hash_value,
                },
            },
            result_parser=resources.ArtifactInstance,
        )

    def list_scans(self, hash_value):
        return PolyswarmRequest(
            self.api_instance,
            {
                'method': 'GET',
                'url': '{}/search/instances'.format(self.uri),
                'params': {
                    'hash': hash_value,
                },
            },
            result_parser=resources.ArtifactInstance,
        )

    def search_metadata(self, query):
        return PolyswarmRequest(
            self.api_instance,
            {
                'method': 'GET',
                'url': '{}/search/metadata/query'.format(self.uri),
                'params': {
                    'query': query,
                },
            },
            result_parser=resources.Metadata,
        )

    def submit(self, artifact, artifact_name, artifact_type):
        return PolyswarmRequest(
            self.api_instance,
            {
                'method': 'POST',
                'url': '{}/consumer/submission/{}'.format(self.uri, self.community),
                'files': {
                    'file': (artifact_name, artifact),
                },
                # very oddly, when included in files parameter this errors out
                'data': {'artifact-type': artifact_type}
            },
            result_parser=resources.ArtifactInstance,
        )

    def rescan(self, hash_value, hash_type):
        return PolyswarmRequest(
            self.api_instance,
            {
                'method': 'POST',
                'url': '{}/consumer/submission/{}/rescan/{}/{}'.format(self.uri, self.community, hash_type, hash_value),
            },
            result_parser=resources.ArtifactInstance,
        )

    def rescanid(self, submission_id):
        return PolyswarmRequest(
            self.api_instance,
            {
                'method': 'POST',
                'url': '{}/consumer/submission/{}/rescan/{}'.format(self.uri, self.community, int(submission_id)),
            },
            result_parser=resources.ArtifactInstance,
        )

    def lookup_uuid(self, submission_id):
        return PolyswarmRequest(
            self.api_instance,
            {
                'method': 'GET',
                'url': '{}/consumer/submission/{}/{}'.format(self.uri, self.community, int(submission_id)),
            },
            result_parser=resources.ArtifactInstance,
        )

    def get_engines(self):
        return PolyswarmRequest(
            self.api_instance,
            {
                'method': 'GET',
                'url': '{}/microengines/list'.format(self.uri),
                'headers': {'Authorization': None},
            },
            result_parser=resources.Engine,
        )

    def create_live_hunt(self, rule=None, rule_id=None, active=True, ruleset_name=None):
        parameters = {
                'method': 'POST',
                'url': '{}/hunt/live'.format(self.uri),
                'json': {'active': active},
            }
        if ruleset_name:
            parameters['json']['ruleset_name'] = ruleset_name
        if rule:
            parameters['json']['yara'] = rule
        if rule_id:
            parameters['json']['rule_id'] = str(int(rule_id))
        return PolyswarmRequest(
            self.api_instance,
            parameters,
            result_parser=resources.Hunt,
        )

    def get_live_hunt(self, hunt_id=None):
        return PolyswarmRequest(
            self.api_instance,
            {
                'method': 'GET',
                'url': '{}/hunt/live'.format(self.uri),
                'params': {'id': str(int(hunt_id)) if hunt_id else ''},
            },
            result_parser=resources.Hunt,
        )

    def update_live_hunt(self, hunt_id=None, active=False):
        return PolyswarmRequest(
            self.api_instance,
            {
                'method': 'PUT',
                'url': '{}/hunt/live'.format(self.uri),
                'params': {'id': str(int(hunt_id)) if hunt_id else ''},
                'json': {'active': active},
            },
            result_parser=resources.Hunt,
        )

    def delete_live_hunt(self, hunt_id):
        return PolyswarmRequest(
            self.api_instance,
            {
                'method': 'DELETE',
                'url': '{}/hunt/live'.format(self.uri),
                'params': {'id': str(int(hunt_id)) if hunt_id else ''},
            },
            result_parser=resources.Hunt,
        )

    def live_list(self, since=None, all_=None):
        parameters = {
            'method': 'GET',
            'url': '{}/hunt/live/list'.format(self.uri),
            'params': {},
        }
        if since is not None:
            parameters['params']['since'] = since
        if all_ is not None:
            parameters['params']['all'] = int(all_)
        return PolyswarmRequest(
            self.api_instance,
            parameters,
            result_parser=resources.Hunt,
        )

    def live_hunt_results(self, hunt_id=None, since=None, tag=None, rule_name=None):
        req = {
            'method': 'GET',
            'url': '{}/hunt/live/results'.format(self.uri),
            'params': {
                'since': since,
                'id': str(int(hunt_id)) if hunt_id else '',
            },
        }
        if tag is not None:
            req['params']['tag'] = tag
        if rule_name is not None:
            req['params']['rule_name'] = rule_name
        return PolyswarmRequest(
            self.api_instance,
            req,
            result_parser=resources.HuntResult,
        )

    def create_historical_hunt(self, rule=None, rule_id=None, ruleset_name=None):
        parameters = {
                'method': 'POST',
                'url': '{}/hunt/historical'.format(self.uri),
                'json': {},
            }
        if ruleset_name:
            parameters['json']['ruleset_name'] = ruleset_name
        if rule:
            parameters['json']['yara'] = rule
        if rule_id:
            parameters['json']['rule_id'] = str(int(rule_id))
        return PolyswarmRequest(
            self.api_instance,
            parameters,
            result_parser=resources.Hunt,
        )

    def get_historical_hunt(self, hunt_id):
        return PolyswarmRequest(
            self.api_instance,
            {
                'method': 'GET',
                'url': '{}/hunt/historical'.format(self.uri),
                'params': {'id': str(int(hunt_id)) if hunt_id else ''},
            },
            result_parser=resources.Hunt,
        )

    def delete_historical_hunt(self, hunt_id):
        return PolyswarmRequest(
            self.api_instance,
            {
                'method': 'DELETE',
                'url': '{}/hunt/historical'.format(self.uri),
                'params': {'id': str(int(hunt_id)) if hunt_id else ''},
            },
            result_parser=resources.Hunt,
        )

    def historical_list(self, since=None):
        parameters = {
            'method': 'GET',
            'url': '{}/hunt/historical/list'.format(self.uri),
            'params': {},
        }
        if since is not None:
            parameters['params']['since'] = since
        return PolyswarmRequest(
            self.api_instance,
            parameters,
            result_parser=resources.Hunt,
        )

    def historical_hunt_results(self, hunt_id=None, tag=None, rule_name=None):
        req = {
            'method': 'GET',
            'url': '{}/hunt/historical/results'.format(self.uri),
            'params': {'id': str(int(hunt_id)) if hunt_id else ''},
        }
        if tag is not None:
            req['params']['tag'] = tag
        if rule_name is not None:
            req['params']['rule_name'] = rule_name
        return PolyswarmRequest(
            self.api_instance,
            req,
            result_parser=resources.HuntResult,
        )

    def create_tag_link(self, sha256, tags=None, families=None):
        parameters = {
            'method': 'POST',
            'url': '{}/tags/link'.format(self.uri),
            'json': {'sha256': sha256},
        }
        if tags:
            parameters['json']['tags'] = tags
        if families:
            parameters['json']['families'] = families
        return PolyswarmRequest(
            self.api_instance,
            parameters,
            result_parser=resources.TagLink,
        )

    def get_tag_link(self, sha256):
        return PolyswarmRequest(
            self.api_instance,
            {
                'method': 'GET',
                'url': '{}/tags/link'.format(self.uri),
                'params': {'hash': sha256},
            },
            result_parser=resources.TagLink,
        )

    def update_tag_link(self, sha256, tags=None, families=None, remove=False):
        parameters = {
            'method': 'PUT',
            'url': '{}/tags/link'.format(self.uri),
            'params': {'hash': sha256},
            'json': {'remove': remove if remove else False},
        }
        if tags:
            parameters['json']['tags'] = tags
        if families:
            parameters['json']['families'] = families
        return PolyswarmRequest(
            self.api_instance,
            parameters,
            result_parser=resources.TagLink,
        )

    def delete_tag_link(self, sha256):
        return PolyswarmRequest(
            self.api_instance,
            {
                'method': 'DELETE',
                'url': '{}/tags/link'.format(self.uri),
                'params': {'hash': sha256},
            },
            result_parser=resources.TagLink,
        )

    def list_tag_link(self, tags=None, families=None, or_tags=None, or_families=None):
        parameters = {
            'method': 'GET',
            'url': '{}/tags/link/list'.format(self.uri),
            'params': [],
        }
        if tags:
            parameters['params'].extend(('tag', p) for p in tags)
        if families:
            parameters['params'].extend(('family', p) for p in families)
        if or_tags:
            parameters['params'].extend(('or_tag', p) for p in or_tags)
        if or_families:
            parameters['params'].extend(('or_family', p) for p in or_families)
        return PolyswarmRequest(
            self.api_instance,
            parameters,
            result_parser=resources.TagLink,
        )

    def create_tag(self, name):
        parameters = {
            'method': 'POST',
            'url': '{}/tags/tag'.format(self.uri),
            'json': {'name': name},
        }
        return PolyswarmRequest(
            self.api_instance,
            parameters,
            result_parser=resources.Tag,
        )

    def get_tag(self, name):
        return PolyswarmRequest(
            self.api_instance,
            {
                'method': 'GET',
                'url': '{}/tags/tag'.format(self.uri),
                'params': {'name': name},
            },
            result_parser=resources.Tag,
        )

    def delete_tag(self, name):
        return PolyswarmRequest(
            self.api_instance,
            {
                'method': 'DELETE',
                'url': '{}/tags/tag'.format(self.uri),
                'params': {'name': name},
            },
            result_parser=resources.Tag,
        )

    def list_tag(self):
        return PolyswarmRequest(
            self.api_instance,
            {
                'method': 'GET',
                'url': '{}/tags/tag/list'.format(self.uri),
            },
            result_parser=resources.Tag,
        )
    
    def create_family(self, name):
        parameters = {
            'method': 'POST',
            'url': '{}/tags/family'.format(self.uri),
            'json': {'name': name},
        }
        return PolyswarmRequest(
            self.api_instance,
            parameters,
            result_parser=resources.MalwareFamily,
        )

    def get_family(self, name):
        return PolyswarmRequest(
            self.api_instance,
            {
                'method': 'GET',
                'url': '{}/tags/family'.format(self.uri),
                'params': {'name': name},
            },
            result_parser=resources.MalwareFamily,
        )

    def delete_family(self, name):
        return PolyswarmRequest(
            self.api_instance,
            {
                'method': 'DELETE',
                'url': '{}/tags/family'.format(self.uri),
                'params': {'name': name},
            },
            result_parser=resources.MalwareFamily,
        )

    def update_family(self, family_name, emerging=True):
        return PolyswarmRequest(
            self.api_instance,
            {
                'method': 'PUT',
                'url': '{}/tags/family'.format(self.uri),
                'params': {'name': family_name},
                'json': {
                    'emerging': emerging if emerging else False
                },
            },
            result_parser=resources.MalwareFamily,
        )

    def list_family(self):
        return PolyswarmRequest(
            self.api_instance,
            {
                'method': 'GET',
                'url': '{}/tags/family/list'.format(self.uri),
            },
            result_parser=resources.MalwareFamily,
        )

    def create_ruleset(self, rule, name, description=None):
        parameters = {
            'method': 'POST',
            'url': '{}/hunt/rule'.format(self.uri),
            'json': {
                'yara': rule,
                'name': name,
            },
        }
        if description:
            parameters['json']['description'] = description
        return PolyswarmRequest(
            self.api_instance,
            parameters,
            result_parser=resources.YaraRuleset,
        )

    def get_ruleset(self, ruleset_id=None):
        return PolyswarmRequest(
            self.api_instance,
            {
                'method': 'GET',
                'url': '{}/hunt/rule'.format(self.uri),
                'params': {'id': str(int(ruleset_id))},
            },
            result_parser=resources.YaraRuleset,
        )

    def update_ruleset(self, ruleset_id, name=None, rules=None, description=None):
        parameters = {
            'method': 'PUT',
            'url': '{}/hunt/rule'.format(self.uri),
            'params': {'id': str(int(ruleset_id))},
            'json': {},
        }
        if name:
            parameters['json']['name'] = name
        if rules:
            parameters['json']['yara'] = rules
        if description:
            parameters['json']['description'] = description
        return PolyswarmRequest(
            self.api_instance,
            parameters,
            result_parser=resources.YaraRuleset,
        )

    def delete_ruleset(self, ruleset_id):
        return PolyswarmRequest(
            self.api_instance,
            {
                'method': 'DELETE',
                'url': '{}/hunt/rule'.format(self.uri),
                'params': {'id': str(int(ruleset_id))},
            },
            result_parser=resources.YaraRuleset,
        )

    def list_ruleset(self):
        return PolyswarmRequest(
            self.api_instance,
            {
                'method': 'GET',
                'url': '{}/hunt/rule/list'.format(self.uri),
            },
            result_parser=resources.YaraRuleset,
        )



class PolyswarmException(Exception):
    pass


#########################################
# API layer exceptions
#########################################

class PolyswarmAPIException(PolyswarmException):
    pass


class TimeoutException(PolyswarmAPIException):
    pass


#########################################
# Request layer exceptions
#########################################

class RequestException(PolyswarmException):
    def __init__(self, request, *args):
        super(PolyswarmException, self).__init__(*args)
        self.request = request


class UsageLimitsExceededException(RequestException):
    pass


class NotFoundException(RequestException):
    pass


class NoResultsException(RequestException):
    pass


#########################################
# Types layer exceptions
#########################################


class TypeException(PolyswarmException):
    pass


class MissingAPIInstanceException(TypeException):
    pass


class InvalidJSONResponseException(TypeException):
    pass


class DecodeErrorException(TypeException):
    pass


class InvalidValueException(TypeException):
    pass


class ArtifactDeletedException(TypeException):
    pass


class InvalidYaraRulesException(TypeException):
    pass


class NotImportedException(TypeException):
    pass



import logging
import time
import os

try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse

from future.utils import string_types

from . import exceptions
from . import const
from . import endpoint
from . import http
from .types import resources


logger = logging.getLogger(__name__)


class PolyswarmAPI(object):
    """A synchronous interface to the public and private PolySwarm APIs."""

    def __init__(self, key, uri=None, community=None, validate_schemas=False, timeout=None):
        """
        :param key: PolySwarm API key
        :param uri: PolySwarm API URI
        :param community: Community to scan against.
        :param validate_schemas: Validate JSON objects when creating response objects. Will impact performance.
        :param timeout: Maximum time to wait for an http response on every request.
        """
        logger.debug('Creating PolyswarmAPI instance: api_key: %s, api_uri: %s, community: %s', key, uri, community)
        self.uri = uri or const.DEFAULT_GLOBAL_API
        self.community = community or const.DEFAULT_COMMUNITY
        self.timeout = timeout or const.DEFAULT_HTTP_TIMEOUT
        self.session = http.PolyswarmHTTP(key, retries=const.DEFAULT_RETRIES)
        self.generator = endpoint.PolyswarmRequestGenerator(self)
        self.validate = validate_schemas
        self._engines = None

    @property
    def engines(self):
        if not self._engines:
            self._engines = self.generator.get_engines().execute().result
            self._engines = {e.address.lower(): e for e in self._engines}
        return self._engines

    def resolve_engine_name(self, eth_pub):
        engine = self.engines.get(eth_pub.lower())
        engine_name = engine.name if engine else eth_pub
        return engine_name.lower()

    def wait_for(self, scan, timeout=const.DEFAULT_SCAN_TIMEOUT):
        """
        Wait for a Scan to scan successfully

        :param scan: Scan id to wait for
        :param timeout: Maximum time in seconds to wait before raising a TimeoutException
        :return: The ArtifactInstance resource waited on
        """
        start = time.time()
        while True:
            scan_result = self.lookup(scan)
            if scan_result.failed or scan_result.window_closed:
                return scan_result
            elif -1 < timeout < time.time() - start:
                raise exceptions.TimeoutException('Timed out waiting for scan {} to finish. Please try again.'
                                                  .format(scan))
            else:
                time.sleep(const.POLL_FREQUENCY)

    def search(self, hash_, hash_type=None):
        """
        Search for the latest scans matching the given hash and hash_type.

        :param hash_: A Hashable object (Artifact, local.LocalArtifact, Hash) or hex-encoded SHA256/SHA1/MD5
        :param hash_type: Hash type of the provided hash_. Will attempt to auto-detect if not explicitly provided.
        :return: Generator of ArtifactInstance resources
        """

        hash_ = resources.Hash.from_hashable(hash_, hash_type=hash_type)
        return self.generator.search_hash(hash_.hash, hash_.hash_type).execute().consume_results()

    def search_scans(self, hash_):
        """
        Search for all scans ever made matching the given sha256.

        :param hash_: A Hashable object (Artifact, local.LocalArtifact, Hash) or hex-encoded SHA256
        :return: Generator of ArtifactInstance resources
        """

        hash_ = resources.Hash.from_hashable(hash_, hash_type='sha256')
        return self.generator.list_scans(hash_.hash).execute().consume_results()

    def search_by_metadata(self, query):
        """
        Search artifacts by metadata

        :param query: A query string
        :return: Generator of ArtifactInstance resources
        """
        return self.generator.search_metadata(query).execute().consume_results()

    def submit(self, artifact, artifact_type=resources.ArtifactType.FILE):
        """
        Submit artifacts to polyswarm and return UUIDs

        :param artifact: A file-like, path to file, url or LocalArtifact instance
        :param artifact_type: The ArtifactType or strings containing "file" or "url"
        :return: An ArtifactInstance resource
        """
        artifact_type = resources.ArtifactType.parse(artifact_type)
        # TODO This is a python 2.7 check if artifact is a file-like instance, consider changing
        #  to isinstance(artifact, io.IOBase) when deprecating 2.7 and implementing making LocalHandle
        #  inherit io.IOBase, although this will change the method delegation logic in the resource
        if hasattr(artifact, 'read') and hasattr(artifact.read, '__call__'):
            artifact = resources.LocalArtifact(artifact, artifact_type=artifact_type, polyswarm=self, analyze=False)
        elif isinstance(artifact, string_types):
            if artifact_type == resources.ArtifactType.FILE:
                artifact = resources.LocalArtifact.from_path(self, artifact, artifact_type=artifact_type)
            elif artifact_type == resources.ArtifactType.URL:
                artifact = resources.LocalArtifact.from_content(self, artifact, artifact_name=artifact,
                                                                artifact_type=artifact_type)
        if isinstance(artifact, resources.LocalArtifact):
            return self.generator.submit(artifact, artifact.artifact_name, artifact.artifact_type.name).execute().result
        else:
            raise exceptions.InvalidValueException('Artifacts should be a path to a file or a LocalArtifact instance')

    def lookup(self, scan):
        """
        Lookup a scan by Scan id.

        :param scan: The Scan UUID to lookup
        :return: An ArtifactInstance resource
        """
        return self.generator.lookup_uuid(scan).execute().result

    def rescan(self, hash_, hash_type=None):
        """
        Rescan a file based on and existing hash in the Polyswarm platform

        :param hash_: Hashable object (Artifact, local.LocalArtifact, or Hash) or hex-encoded SHA256/SHA1/MD5
        :param hash_type: Hash type of the provided hash_. Will attempt to auto-detect if not explicitly provided.
        :return: A ArtifactInstance resources
        """
        hash_ = resources.Hash.from_hashable(hash_, hash_type=hash_type)
        return self.generator.rescan(hash_.hash, hash_.hash_type).execute().result

    def rescan_id(self, scan):
        """
        Re-execute a new scan based on an existing scan.

        :param scan: Id of the existing scan
        :return: A ArtifactInstance resource
        """
        return self.generator.rescanid(scan).execute().result

    def _parse_rule(self, rule):
        if isinstance(rule, string_types):
            rule, rule_id = resources.YaraRuleset(dict(yara=rule), polyswarm=self), None
            try:
                rule.validate()
            except exceptions.NotImportedException as e:
                logger.debug('%s\nSkipping validation.', str(e))
        elif isinstance(rule, (resources.YaraRuleset, int)):
            rule, rule_id = None, rule
        else:
            raise exceptions.InvalidValueException('Either yara or rule_id must be provided.')
        return rule, rule_id

    def live_create(self, rule, active=True, ruleset_name=None):
        """
        Create a new live hunt_id, and replace the currently running YARA rules.

        :param rule: YaraRuleset object or string containing YARA rules to install
        :param active: Set the live hunt to active upon creation if True.
        :param ruleset_name: Name of the ruleset.
        :return: The created Hunt resource
        """
        rule, rule_id = self._parse_rule(rule)
        return self.generator.create_live_hunt(rule=rule.yara if rule else None, rule_id=rule_id,
                                               active=active, ruleset_name=ruleset_name).execute().result

    def live_get(self, hunt=None):
        """
        Delete a live hunt.

        :param hunt: Hunt ID
        :return: The Hunt resource
        """
        return self.generator.get_live_hunt(hunt).execute().result

    def live_update(self, active, hunt=None):
        """
        Update a live hunt.

        :param hunt: Hunt ID
        :param active: True to start the live hunt and False to stop it
        :return: The updated Hunt resource
        """
        return self.generator.update_live_hunt(hunt, active=active).execute().result

    def live_delete(self, hunt=None):
        """
        Delete a live hunt.

        :param hunt: Hunt ID
        :return: The deleted Hunt resource
        """
        return self.generator.delete_live_hunt(hunt).execute().result

    def live_list(self, since=None, all_=None):
        """
        List all the live hunts

        :return: Generator of Hunt resources
        """
        return self.generator.live_list(since=since, all_=all_).execute().consume_results()

    def live_results(self, hunt=None, since=None, tag=None, rule_name=None):
        """
        Get results from a live hunt

        :param hunt: ID of the hunt (None if results for tha latest active hunt are desired)
        :param since: Fetch results from the last "since" minutes
        :param tag: Filter hunt results containing the provided tags (comma separated tags, exact match).
        :param rule_name: Filter hunt results on the provided rule name (exact match).
        :return: Generator of HuntResult resources
        """
        return self.generator.live_hunt_results(hunt_id=hunt, since=since,
                                                tag=tag, rule_name=rule_name).execute().consume_results()

    def historical_create(self, rule=None, ruleset_name=None):
        """
        Run a new historical hunt.

        :param rule: YaraRuleset object or string containing YARA rules to install
        :param ruleset_name: Name of the ruleset.
        :return: The created Hunt resource
        """
        rule, rule_id = self._parse_rule(rule)
        return self.generator.create_historical_hunt(rule=rule.yara if rule else None, rule_id=rule_id,
                                                     ruleset_name=ruleset_name).execute().result

    def historical_get(self, hunt=None):
        """
        Delete a live hunt.

        :param hunt: Hunt ID
        :return: The Hunt resource
        """
        return self.generator.get_historical_hunt(hunt).execute().result

    def historical_delete(self, hunt):
        """
        Delete a historical hunts.

        :param hunt: Hunt ID
        :return: The deleted Hunt resource
        """
        return self.generator.delete_historical_hunt(hunt).execute().result

    def historical_list(self, since=None):
        """
        List all historical hunts

        :return: Generator of Hunt resources
        """
        return self.generator.historical_list(since=since).execute().consume_results()

    def historical_results(self, hunt=None, tag=None, rule_name=None):
        """
        Get results from a historical hunt

        :param hunt: ID of the hunt (None if latest hunt results are desired)
        :param tag: Filter hunt results containing the provided tags (comma separated tags, exact match).
        :param rule_name: Filter hunt results on the provided rule name (exact match).
        :return: Generator of HuntResult resources
        """
        return self.generator.historical_hunt_results(hunt_id=hunt, tag=tag, rule_name=rule_name).execute().consume_results()

    def ruleset_create(self, name, rules, description=None):
        """
        Create a Yara Ruleset from the provided rules with the given name in the polyswarm platform.
        :param name: Name of the ruleset
        :param rules: Yara rules as a string
        :param description: Description of the ruleset
        :return: A YaraRuleset resource
        """
        rules = resources.YaraRuleset(dict(name=name, description=description, yara=rules), polyswarm=self)
        try:
            rules.validate()
        except exceptions.NotImportedException as e:
            logger.debug('%s\nSkipping validation.', str(e))
        return self.generator.create_ruleset(rules.yara, rules.name, description=rules.description).execute().result

    def ruleset_get(self, ruleset_id=None):
        """
        Retrieve a YaraRuleset from the polyswarm platform by its Id.
        :param ruleset_id: Id of the ruleset
        :return: A YaraRuleset resource
        """
        return self.generator.get_ruleset(ruleset_id).execute().result

    def ruleset_update(self, ruleset_id, name=None, rules=None, description=None):
        """
        Update an existing YaraRuleset in the polyswarm platform by its Id.
        :param ruleset_id: Id of the ruleset
        :param name: New name of the ruleset
        :param rules: New yara rules as a string
        :param description: New description of the ruleset
        :return: The updated YaraRuleset resource
        """
        return self.generator.update_ruleset(ruleset_id, name=name, rules=rules, description=description).execute().result

    def ruleset_delete(self, ruleset_id):
        """
        Delete a YaraRuleset from the polyswarm platform by its Id.
        :param ruleset_id: Id of the ruleset
        :return: A YaraRuleset resource
        """
        return self.generator.delete_ruleset(ruleset_id).execute().result

    def ruleset_list(self):
        """
        List all YaraRulesets for the current account.
        :return: A generator of YaraRuleset resources
        """
        return self.generator.list_ruleset().execute().consume_results()

    def tag_link_get(self, sha256):
        """
        Fetch the Tags and Families associated with the given sha256.

        :param sha256: The sha256 of the artifact.
        :return: A TagLink resource
        """
        return self.generator.get_tag_link(sha256).execute().result

    def tag_link_update(self, sha256, tags=None, families=None, remove=False):
        """
        Update a TagLink with the given type or value by its id.
        :param sha256: The sha256 of the artifact.
        :param tags: A list of tags to be added or removed.
        :param families: A list of families to be added or removed.
        :param remove: A flag indicating if we should remove the provided tags/families.
        :return: A TagLink resource
        """
        return self.generator.update_tag_link(sha256, tags=tags, families=families, remove=remove).execute().result

    def tag_link_list(self, tags=None, families=None, or_tags=None, or_families=None):
        """
        Fetch all existing TagLinks for the provided tags.
        :param tags: A list of tags that must be associated with the TagLinks listed.
        :param families: A list of families that must be associated with the TagLinks listed.
        :param or_tags: A list of tags that the TagLinks must be associated with at least one.
        :param or_families: A list of families that the TagLinks must be associated with at least one.
        :return: A TagLink resource
        """
        return self.generator.list_tag_link(tags=tags, families=families,
                                            or_tags=or_tags, or_families=or_families).execute().consume_results()

    def tag_create(self, name):
        """
        Create a Tag.
        :param name: The tag we want to create.
        :return: A Tag resource
        """
        return self.generator.create_tag(name).execute().result

    def tag_get(self, name):
        """
        Fetch a Tag.
        :param name: The tag we want to fetch.
        :return: A Tag resource
        """
        return self.generator.get_tag(name).execute().result

    def tag_delete(self, name):
        """
        Delete a Tag.
        :param name: The tag we want to delete.
        :return: A Tag resource
        """
        return self.generator.delete_tag(name).execute().result

    def tag_list(self):
        """
        Fetch all existing Tags.
        :return: A generator of Tag resources
        """
        return self.generator.list_tag().execute().consume_results()

    def family_create(self, name):
        """
        Create a Family.
        :param name: The family name.
        :return: A MalwareFamily resource
        """
        return self.generator.create_family(name).execute().result

    def family_get(self, name):
        """
        Fetch a Family.
        :param name: The family name.
        :return: A MalwareFamily resource
        """
        return self.generator.get_family(name).execute().result

    def family_delete(self, name):
        """
        Delete a Family.
        :param name: The family name.
        :return: A MalwareFamily resource
        """
        return self.generator.delete_family(name).execute().result

    def family_update(self, family_name, emerging=True):
        """
        Update the Family emerging status.
        :param family_name: The family name.
        :param emerging: A flag indicating if the family should be marked as emerging at this point in time.
        :return: A MalwareFamily resource
        """
        return self.generator.update_family(family_name, emerging=emerging).execute().result

    def family_list(self):
        """
        Fetch all existing Families
        :return: A generator of MalwareFamily resources
        """
        return self.generator.list_family().execute().consume_results()

    def download(self, out_dir, hash_, hash_type=None):
        """
        Grab the data of artifact identified by hash, and write the data to a file in the provided directory
        under a file named after the hash_.
        :param out_dir: Destination directory to download the file.
        :param hash_: The hash we should use to lookup the artifact to download.
        :param hash_type: Hash type of the provided hash_. Will attempt to auto-detect if not explicitly provided.
        :return: A LocalArtifact resource
        """
        hash_ = resources.Hash.from_hashable(hash_, hash_type=hash_type)
        path = os.path.join(out_dir, hash_.hash)
        artifact = resources.LocalArtifact.from_path(self, path, create=True)
        self.generator.download(hash_.hash, hash_.hash_type, handle=artifact).execute()
        return artifact

    def download_archive(self, out_dir, s3_path):
        """
        Grab the data in the s3 path provided in the stream() method, and write the contents
        in the provided directory.
        :param out_dir: Destination directory to download the file.
        :param s3_path: Target S3 object to download.
        :return: A LocalArtifact resource
        """
        path = os.path.join(out_dir, os.path.basename(urlparse(s3_path).path))
        artifact = resources.LocalArtifact.from_path(self, path, create=True)
        self.generator.download_archive(s3_path, handle=artifact).execute()
        return artifact

    def download_to_handle(self, hash_, fh, hash_type=None):
        """
        Grab the data of artifact identified by hash, and write the data to a file handle
        :param hash_: The hash we should use to lookup the artifact to download.
        :param fh: A file-like object which we are going to write the contents of the artifact to.
        :param hash_type: Hash type of the provided hash_. Will attempt to auto-detect if not explicitly provided.
        :return: A LocalHandle resource
        """
        hash_ = resources.Hash.from_hashable(hash_, hash_type=hash_type)
        return self.generator.download(hash_.hash, hash_.hash_type, handle=fh).execute().result

    def stream(self, since=const.MAX_SINCE_TIME_STREAM):
        """
        Access the stream of artifacts (ask info@polyswarm.io about access)

        :param since: Fetch results from the last "since" minutes (up to 2 days)
        :return: Generator of ArtifactArchive resources
        """
        return self.generator.stream(since=since).execute().consume_results()



artifact_instance_schema = {
    'type': 'object',
    'properties': {
        'artifact_id': {'type': 'string'},
        'bounty_id': {'type': ['null', 'string']},
        'bounty_result': {
            'oneOf': [
                {'type': 'null'},
                {
                    'type': 'object',
                    'properties': {
                        'artifact_type': {'type': 'string'},
                        'files': {'type': 'array'}, # has custom validation, we should flatten this in AI
                        'permalink': {'type': 'string'},
                        'status': {'type': 'string'},
                        'uuid': {'type': 'string'},
                    },
                },
            ],
        },
        'community': {'type': 'string'},
        'consumer_guid': {'type': ['string', 'null']},
        'country': {'type': ['string', 'null']},
        'id': {'type': 'string'},
        'name': {'type': ['string', 'null']},
        'submitted': {'type': 'string'},
    },
    'required': ['artifact_id', 'bounty_id', 'bounty_result', 'community', 'country', 'id', 'name', 'submitted']
}

artifact_schema = {
    'type': 'object',
    'properties': {
        'artifact_instances': {'type': 'array'},
        'artifact_metadata': {'type': 'object'},
        'extended_type': {'type': 'string', 'minLength': 1},
        'first_seen': {'type': 'string'},
        'id': {'type': 'string'},
        'sha256': {
            'type': 'string',
            'minLength': 64,
            'maxLength': 64,
        },
        'sha1': {
            'type': 'string',
            'minLength': 40,
            'maxLength': 40,
        },
        'md5': {
            'type': 'string',
            'minLength': 32,
            'maxLength': 32,
        },
        'mimetype': {
            'type': 'string',
            'minLength': 1,
        },
    },
    'required': ['extended_type', 'first_seen', 'id', 'sha256', 'sha1', 'md5', 'mimetype'],
}

artifact_archive_schema = {
    'type': 'object',
    'properties': {
        'id': {'type': 'string'},
        's3_path': {
            'type': 'string',
        },
        'community': {
            'type': 'string',
        },
        'created': {'type': 'string'},

    },
    'required': ['s3_path'],
}

# TODO fill out more
artifact_metadata = {
    'type': 'object',
}



api_response_schema = {
    'type': 'object',
    'properties': {
        'status': {'type': 'string'},
        'result': {'type': ['array', 'object', 'string']},
        'errors': {'type': 'string'},
        'total': {'type': 'integer'},
        'limit': {'type': 'integer'},
        'page': {'type': 'integer'},
        'orderBy': {'type': 'string'},
        'direction': {'type': 'string'}
    },
    'required': ['status', 'result'],
}


#!/usr/bin/env python
# -*- coding: utf-8 -*-
from setuptools import find_packages, setup

# The README.md will be used as the content for the PyPi package details page on the Python Package Index.
with open('README.md', 'r') as readme:
    long_description = readme.read()


setup(
    name='polyswarm-api',
    version='2.0.3.dev2',
    description='Client library to simplify interacting with the PolySwarm consumer API',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='PolySwarm Developers',
    author_email='info@polyswarm.io',
    url='https://github.com/polyswarm/polyswarm-api',
    license='MIT',
    python_requires='>=2.7,<4',
    install_requires=[
        'requests~=2.22.0',
        'jsonschema~=3.0.2',
        'ordered-set~=3.1.1',
        'future~=0.18.2',
        'python-dateutil~=2.8.1',
    ],
    extras_require={':python_version < "3.0"': ['futures==3.3.0', 'enum34==1.1.6']},
    include_package_data=True,
    packages=find_packages('src'),
    package_dir={'': 'src'},
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: PyPy',
    ]
)




# TODO change API so that metadata isn't a string when empty...and also not a list
# TODO much of this could be taken from polyswarm-artifact
assertion_schema = {
    'type': 'object',
    'properties': {
        'author': {'type': 'string'},
        'bid': {'type': ['string', 'integer']},
        'mask': {'type': 'boolean'},
        'metadata': {
            'oneOf': [
                {'type': 'string'},
                {'type': 'array'},
                {
                    'type': 'object',
                    'properties': {
                        'malware_family': {'type': 'string'},
                        'scanner': {
                            'type': 'object',
                            'properties': {
                                'environment': {'type': 'object'},
                            }
                        }
                    },
                    'required': ['malware_family']
                },
            ]
        },
        'verdict': {'type': ['boolean', 'null']},
    },
    'required': ['author', 'bid', 'mask', 'metadata', 'verdict']
}

vote_schema = {
    'type': 'object',
    'properties': {
        'arbiter': {'type': 'string'},
        'vote': {'type': 'boolean'},
    },
    'required': ['arbiter', 'vote']
}

bounty_file_schema = {
    'type': 'object',
    'properties': {
        'assertions': {
            'type': 'array',
            'items': {
                'type': 'object'
            }
        },
        'bounty_guid': {'type': ['string', 'null']},
        'bounty_status': {'type': ['string', 'null']},
        'failed': {'type': 'boolean'},
        'filename': {'type': 'string'},
        'hash': {'type': 'string'},
        'result': {'type': ['boolean', 'null']},
        'size': {'type': 'integer'},
        'votes': {
            'type': 'array',
            'items': {
                'type': 'object'
            }
        },
        'window_closed': {'type': 'boolean'},
        'id': {'type': 'string'},
    },
    'required': ['assertions', 'bounty_guid', 'bounty_status', 'failed', 'filename', 'hash', 'result',
                 'size', 'votes', 'window_closed']
}

bounty_schema = {
    'type': 'object',
    'properties': {
        'artifact_type': {'type': 'string'},
        'files': {
            'type': 'array',
            'items': {
                'type': 'object'
            },
        },  # has custom validation, we should flatten this in AI
        'permalink': {'type': 'string'},
        'status': {'type': 'string'},
        'uuid': {'type': 'string'},
    },
    'required': ['files', 'status', 'uuid'],
}

polyscore_schema = {
    'type': 'object',
    'properties': {
        'scores': {'type': 'object'},
    },
    'required': ['scores']
}
